"""
Large default configuration values are stored here
"""
import yaml

SELF_SERVICE_IAM_DEFAULTS = yaml.safe_load(
    """
permissions_map:
  s3:
    text: S3 Bucket
    description: S3 Permissions
    inputs:
      - name: resource_arn
        type: typeahead_input
        text: Bucket Name
        required: true
        typeahead_endpoint: /policies/typeahead?resource=s3&show_full_arn_for_s3_buckets=true&search={query}
      - name: bucket_prefix
        type: text_input
        text: Prefix (Folder under S3 that you need access to).
        required: true
        default: /*
    action_map:
      - name: list
        text: List
        permissions:
          - s3:ListBucket
          - s3:ListBucketVersions
      - name: get
        text: Get
        permissions:
          - s3:GetObject
          - s3:GetObjectTagging
          - s3:GetObjectVersion
          - s3:GetObjectVersionTagging
          - s3:GetObjectAcl
          - s3:GetObjectVersionAcl
      - name: put
        text: Put
        permissions:
          - s3:PutObject
          - s3:PutObjectTagging
          - s3:PutObjectVersionTagging
          - s3:ListMultipartUploadParts*
          - s3:AbortMultipartUpload
      - name: delete
        text: Delete
        permissions:
          - s3:DeleteObject
          - s3:DeleteObjectTagging
          - s3:DeleteObjectVersion
          - s3:DeleteObjectVersionTagging
  sqs:
    text: SQS Queue
    description: ""
    inputs:
      - name: resource_arn
        type: typeahead_input
        text: Queue ARN
        required: true
        typeahead_endpoint: /policies/typeahead?resource=sqs&search={query}
    action_map:
      - name: send_messages
        text: Send Message (Queue Producer)
        permissions:
          - sqs:GetQueueAttributes
          - sqs:GetQueueUrl
          - sqs:SendMessage
      - name: receive_messages
        text: Receive/Delete Messages (Queue Consumer)
        permissions:
          - sqs:GetQueueAttributes
          - sqs:GetQueueUrl
          - sqs:ReceiveMessage
          - sqs:DeleteMessage
      - name: set_queue_attributes
        text: Set Queue Attributes
        permissions:
          - sqs:SetQueueAttributes
      - name: purge_messages
        text: Purge Queue (You monster!)
        permissions:
          - sqs:PurgeQueue
  sns:
    text: SNS Topic
    description: ""
    inputs:
      - name: resource_arn
        type: typeahead_input
        text: Topic ARN
        required: true
        typeahead_endpoint: /policies/typeahead?resource=sns&search={query}
    action_map:
      - name: get_topic_attributes
        text: Get Topic Attributes
        permissions:
          - sns:GetEndpointAttributes
          - sns:GetTopicAttributes
      - name: publish
        text: Publish
        permissions:
          - sns:Publish
      - name: subscribe
        text: Subscribe
        permissions:
          - sns:Subscribe
          - sns:ConfirmSubscription
      - name: unsubscribe
        text: Unsubscribe
        permissions:
          - sns:Unsubscribe
  rds:
    text: RDS
    description: ""
    inputs:
      - name: resource_arn
        type: text_input
        text: EC2 resource you need access to
        required: true
        default: "arn:aws:iam::{account_id}:role/rds-monitoring-role"
    action_map:
      - name: passrole
        text: PassRole
        permissions:
          - iam:PassRole
  ec2:
    text: EC2
    description: ""
    inputs:
      - name: resource_arn
        type: text_input
        text: EC2 resource you need access to
        required: true
        default: "*"
    action_map:
      - name: volmount
        text: VolMount
        permissions:
          - ec2:attachvolume
          - ec2:createvolume
          - ec2:describelicenses
          - ec2:describevolumes
          - ec2:detachvolume
          - ec2:reportinstancestatus
          - ec2:resetsnapshotattribute
      - name: ipv6
        text: Enable IPv6
        permissions:
          - ec2:AssignIpv6Addresses
  route53:
    text: Route53
    description: ""
    inputs:
      - name: resource_arn
        type: text_input
        text: Route53 Domain you need access to
        required: true
        default: "*"
    action_map:
      - name: list_records
        text: List Records
        permissions:
          - route53:listresourcerecordsets
      - name: change_records
        text: Change Records
        permissions:
          - route53:changeresourcerecordsets
  sts:
    text: STS AssumeRole
    description: ""
    inputs:
      - name: resource_arn
        type: typeahead_input
        text: Role ARN that you wish to assume
        required: true
        typeahead_endpoint: /policies/typeahead?resource=iam_arn&search={query}
    action_map:
      - name: assume_role
        text: Assume Role
        permissions:
          - sts:AssumeRole
  ses:
    text: SES - Send Email
    inputs:
      - name: from_address
        type: text_input
        text: Email Address to send from
        required: true
        default: "emailaddress@example.com"
      - name: resource_arn
        type: text_input
        text: ARN of the resource to send from
        required: true
    condition:
      StringLike:
        ses:FromAddress: "${from_address}"
    action_map:
      - name: send_email
        text: Send Email
        permissions:
          - ses:SendEmail
          - ses:SendRawEmail
  crud_lookup:
    text: Other
    description: |-
      Define the service (rds, route53, rekognition, etc), the resource (An ARN or a wildcard), and the list of actions
      (List, Read, Write, Permissions Management, Tagging) that you need access to. If you know the specific IAM
      permissions you need, use "Advanced Mode" from the dropdown instead.
    inputs:
      - name: service_name
        type: single_typeahead_input
        text: Service (s3, sqs, sns, rekognition, etc)
        required: true
        typeahead_endpoint: /api/v1/policyuniverse/autocomplete/?only_filter_services=true&prefix={query}
      - name: resource_arn
        type: typeahead_input
        text: Resource ARN or Wildcard (*)
        required: true
        typeahead_endpoint: /api/v2/typeahead/resources?typeahead={query}&ui_formatted=true
    action_map:
      - name: list
        text: List
        permissions:
          - List
      - name: read
        text: Read
        permissions:
          - Read
      - name: write
        text: Write
        permissions:
          - Write
      - name: permissions-management
        text: Permissions Management
        permissions:
          - Permissions management
      - name: tagging
        text: Tagging
        permissions:
          - Tagging"""
)

PERMISSION_TEMPLATE_DEFAULTS = yaml.safe_load(
    """
- key: default
  text: Default Template
  value: |-
    {
         "Statement":[
             {
                 "Action":[
                     ""
                 ],
                 "Effect":"Allow",
                 "Resource": [
                     ""
                 ]
             }
         ],
         "Version":"2012-10-17"
     }
- key: s3write
  text: S3 Write Access
  value: |-
    {
        "Statement":[
            {
                "Action":[
                    "s3:ListBucket",
                    "s3:GetObject",
                    "s3:PutObject",
                    "s3:DeleteObject"
                ],
                "Effect":"Allow",
                "Resource":[
                    "arn:aws:s3:::BUCKET_NAME",
                    "arn:aws:s3:::BUCKET_NAME/OPTIONAL_PREFIX/*"
                ],
                "Sid":"s3readwrite"
            }
        ]
    }
- key: s3read
  text: S3 Read Access
  value: |-
    {
        "Statement":[
            {
                "Action":[
                    "s3:ListBucket",
                    "s3:GetObject"
                ],
                "Effect":"Allow",
                "Resource":[
                    "arn:aws:s3:::BUCKET_NAME",
                    "arn:aws:s3:::BUCKET_NAME/OPTIONAL_PREFIX/*"
                ],
                "Sid":"s3readonly"
            }
        ]
    }
- key: sqs
  text: SQS
  value: |-
    {
      "Statement": [
        {
          "Action": [
            "sqs:ReceiveMessage",
            "sqs:SendMessage",
            "sqs:DeleteMessage",
            "sqs:GetQueueUrl",
            "sqs:GetQueueAttributes"
          ],
          "Effect": "Allow",
          "Resource": "QUEUE_ARN"
        }
      ]
    }
- key: sns
  text: SNS
  value: |-
    {
      "Statement": [
        {
          "Action": [
            "sns:Publish",
            "sns:Subscribe",
            "sns:Unsubscribe"
          ],
          "Effect": "Allow",
          "Resource": "TOPIC_ARN"
        }
      ]
    }
- key: rdspassrole
  text: RDS Monitoring PassRole
  value: |-
    {
      "Statement": [
        {
          "Action": [
            "iam:PassRole"
          ],
          "Effect": "Allow",
          "Resource": "arn:aws:iam::<ACCOUNTID>:role/rds-monitoring-role"
        }
      ]
    }
- key: ses
  text: SES
  value: |-
    {
        "Statement": [
            {
                "Action": [
                    "ses:SendEmail",
                    "ses:SendRawEmail"
                ],
                "Effect": "Allow",
                "Resource": "arn:aws:ses:*:123456789012:identity/example.com",
                "Condition": {
                    "StringLike": {
                        "ses:FromAddress": "SENDER@example.com"
                    }
                }
            }
        ]
    }
- key: sts
  text: STS - Assume role
  value: |-
    {
        "Statement": [
            {
                "Action": [
                    "sts:AssumeRole"
                ],
                "Effect": "Allow",
                "Resource": "arn:aws:iam::ACCOUNT_NUMBER:role/ROLE"
            }
        ]
    }
- key: route53
  text: Route53 List/Change Record Sets
  value: |-
    {
        "Statement": [
            {
                "Action": [
                    "route53:changeresourcerecordsets",
                    "route53:listresourcerecordsets"
                ],
                "Resource": [
                    "*"
                ],
                "Effect": "Allow"
            }
        ]
    }
- key: ec2_create_volume
  text: EC2 Volmount
  value: |-
    {
        "Statement": [
            {
                "Action": [
                    "ec2:attachvolume",
                    "ec2:createvolume",
                    "ec2:describelicenses",
                    "ec2:describevolumes",
                    "ec2:detachvolume",
                    "ec2:reportinstancestatus",
                    "ec2:resetsnapshotattribute"
                ],
                "Resource": [
                    "*"
                ],
                "Effect": "Allow"
            }
        ]
    }
- key: put_cloudwatch_data
  text: Cloudwatch - putmetricdata
  value: |-
    {
        "Statement":[
            {
                "Action":[
                    "cloudwatch:putmetricdata"
                ],
                "Effect":"Allow",
                "Resource":[
                    "*"
                ]
            }
        ],
        "Version":"2012-10-17"
    }
- key: eni_auto_attach
  text: ENI - Auto attach
  value: |-
    {
        "Statement":[
            {
                "Action":[
                    "ec2:AttachNetworkInterface",
                    "ec2:Describe*",
                    "ec2:DetachNetworkInterface"
                ],
                "Effect":"Allow",
                "Resource":[
                    "*"
                ],
                "Sid":"eniauto"
            }
        ],
        "Version":"2012-10-17"
    }
- key: ec2_ipv6
  text: EC2 - AssignIpv6Addresses
  value: |-
    {
        "Statement":[
            {
                "Action":[
                    "ec2:AssignIpv6Addresses"
                ],
                "Effect":"Allow",
                "Resource":[
                    "*"
                ]
            }
        ]
    }
- key: volmount
  text: EC2 - volmount
  value: |-
    {
        "Statement":[
            {
                "Action":[
                      "ec2:attachvolume",
                      "ec2:createsnapshot",
                      "ec2:createtags",
                      "ec2:createvolume",
                      "ec2:deletesnapshot",
                      "ec2:describeinstances",
                      "ec2:describetags",
                      "ec2:describevolumes",
                      "ec2:modifyinstanceattribute"
                ],
                "Effect":"Allow",
                "Resource":[
                    "*"
                ]
            }
        ]
    }
-   key: securitygroupmutate
    text: EC2 - Security Group Read/Write
    value: |-
      {
        "Statement": [
            {
                "Action": [
                    "ec2:authorizesecuritygroupingress",
                    "ec2:revokesecuritygroupingress"
                ],
                "Effect": "Allow",
                "Resource": [
                    "arn:aws:ec2:REGION:ACCOUNT_NUMBER:security-group/SECURITY_GROUP_ID"
                ],
                "Sid": "sgmutate"
            },
            {
                "Action": [
                    "ec2:describesecuritygroup*"
                ],
                "Effect": "Allow",
                "Resource": [
                    "*"
                ],
                "Sid": "sgdescribe"
            }
        ]
      }
"""
)



import sys

import boto3
import ujson as json
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError

from consoleme.config import config
from consoleme.lib.plugins import get_plugin_by_name

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


async def duo_mfa_user(username, message="ConsoleMe Authorization Request"):
    """Send a DUO mfa push request to user."""
    # Create session for the region deployed in.
    session = boto3.Session(region_name=config.region)

    # Create lambda client
    client = session.client("lambda")

    # Generate the payload for the event passed to the lambda function
    payload = {"username": username, "message_type": message}

    lambda_arn = config.get("duo.lambda_arn", None)

    log_data = {"function": f"{__name__}.{sys._getframe().f_code.co_name}"}

    if lambda_arn:
        try:
            # Invoke the Lambda Function that will send a DUO Push to the user
            response = await sync_to_async(client.invoke)(
                FunctionName=lambda_arn.format(config.region),
                InvocationType="RequestResponse",
                Payload=bytes(json.dumps(payload), "utf-8"),
            )

            stats.count("duo.mfa_request", tags={"user": username})
        except ClientError as e:
            log_data["error"] = e.response.get("Error", {}).get(
                "Message", "Unknown error in Duo Lambda invoke"
            )
            log.error(log_data, exc_info=True)
            # We had an error so we should deny this request
            return False

        log_data["message"] = "Duo MFA request sent to {}".format(username)

        log.info(log_data)

        # Decode and return the result
        return await decode_duo_response_from_lambda(response, username)


async def decode_duo_response_from_lambda(response, username):
    """Decode the response from the Duo lambda."""
    result = json.loads(response["Payload"].read().decode("utf-8"))
    if not result:
        return False
    if result.get("duo_auth", "") == "success":
        stats.count("duo.mfa_request.approve", tags={"user": username})
        return True
    stats.count("duo.mfa_request.deny", tags={"user": username})
    return False



import os
import shutil
import tempfile
from typing import Optional

import git
from asgiref.sync import sync_to_async


class Repository:
    def __init__(self, repo_url, repo_name, git_email):
        self.tempdir = tempfile.mkdtemp()
        self.repo_url = repo_url
        self.git_email = git_email
        self.repo = None
        self.repo_name = repo_name
        self.git = None

    async def clone(self, no_checkout=True, depth: Optional[int] = None):
        args = []
        kwargs = {}
        if no_checkout:
            args.append("-n")
        args.append(self.repo_url)
        if depth:
            kwargs["depth"] = depth
        await sync_to_async(git.Git(self.tempdir).clone)(*args, **kwargs)
        self.repo = git.Repo(os.path.join(self.tempdir, self.repo_name))
        self.repo.config_writer().set_value("user", "name", "ConsoleMe").release()
        if self.git_email:
            self.repo.config_writer().set_value(
                "user", "email", self.git_email
            ).release()
        self.git = self.repo.git
        return self.repo

    async def cleanup(self):
        await sync_to_async(shutil.rmtree)(self.tempdir)



import ujson as json
from mock import patch
from tornado.testing import AsyncHTTPTestCase

from tests.conftest import MockBaseHandler, MockBaseMtlsHandler, create_future


class TestRolesHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        self.config = config
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_get(self):
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch("/api/v2/roles", method="GET", headers=headers)
        self.assertEqual(response.code, 200)
        responseJSON = json.loads(response.body)
        self.assertIn("eligible_roles", responseJSON)
        self.assertEqual(0, len(responseJSON["eligible_roles"]))

    @patch(
        "consoleme.handlers.v2.roles.RolesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_create_unauthorized_user(self):
        expected = {
            "status": 403,
            "title": "Forbidden",
            "message": "User is unauthorized to create a role",
        }
        response = self.fetch("/api/v2/roles", method="POST", body="test")
        self.assertEqual(response.code, 403)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RolesHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.handlers.v2.roles.can_create_roles")
    def test_create_authorized_user(self, mock_can_create_roles):
        mock_can_create_roles.return_value = True
        input_body = {
            "account_id": "012345678901",
            "description": "This description should be added",
            "instance_profile": "True",
        }
        expected = {
            "status": 400,
            "title": "Bad Request",
            "message": "Error validating input: 1 validation error for RoleCreationRequestModel\nRoleName\n"
            "  field required (type=value_error.missing)",
        }
        response = self.fetch(
            "/api/v2/roles", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 400)
        self.assertDictEqual(json.loads(response.body), expected)

        input_body["role_name"] = "fakeRole"
        expected = {
            "errors": 0,
            "role_created": "true",
            "action_results": [
                {
                    "status": "success",
                    "message": "Role arn:aws:iam::012345678901:role/fakeRole successfully created",
                },
                {
                    "status": "success",
                    "message": "Successfully added default Assume Role Policy Document",
                },
                {
                    "status": "success",
                    "message": "Successfully added description: This description should be added",
                },
                {
                    "status": "success",
                    "message": "Successfully added instance profile fakeRole to role fakeRole",
                },
            ],
        }
        response = self.fetch(
            "/api/v2/roles", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        self.assertDictEqual(json.loads(response.body), expected)


class TestAccountRolesHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        self.config = config
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_get(self):
        expected = {
            "status": 501,
            "title": "Not Implemented",
            "message": "Get roles by account",
        }
        headers = {
            self.config.get("auth.user_header_name"): "user@github.com",
            self.config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }
        response = self.fetch(
            "/api/v2/roles/012345678901", method="GET", headers=headers
        )
        self.assertEqual(response.code, 501)
        self.assertDictEqual(json.loads(response.body), expected)


class TestRoleDetailHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    @patch("consoleme.handlers.v2.roles.RoleDetailHandler.authorization_flow")
    def test_delete_no_user(self, mock_auth):
        mock_auth.return_value = create_future(None)
        expected = {"status": 403, "title": "Forbidden", "message": "No user detected"}
        response = self.fetch(
            "/api/v2/roles/012345678901/fake_account_admin", method="DELETE"
        )
        self.assertEqual(response.code, 403)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RoleDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_delete_unauthorized_user(self):
        expected = {
            "status": 403,
            "title": "Forbidden",
            "message": "User is unauthorized to delete a role",
        }
        response = self.fetch(
            "/api/v2/roles/012345678901/fake_account_admin", method="DELETE"
        )
        self.assertEqual(response.code, 403)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RoleDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.handlers.v2.roles.can_delete_iam_principals")
    def test_delete_authorized_user_invalid_role(self, mock_can_delete_iam_principals):
        expected = {
            "status": 500,
            "title": "Internal Server Error",
            "message": "Error occurred deleting role: An error occurred (NoSuchEntity) when calling the GetRole "
            "operation: Role fake_account_admin not found",
        }
        mock_can_delete_iam_principals.return_value = True
        response = self.fetch(
            "/api/v2/roles/012345678901/fake_account_admin", method="DELETE"
        )
        self.assertEqual(response.code, 500)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RoleDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.handlers.v2.roles.can_delete_iam_principals")
    def test_delete_authorized_user_valid_role(self, mock_can_delete_iam_principals):
        import boto3

        from consoleme.config import config

        client = boto3.client(
            "iam", region_name="us-east-1", **config.get("boto3.client_kwargs", {})
        )
        role_name = "fake_account_admin"
        account_id = "123456789012"
        client.create_role(RoleName=role_name, AssumeRolePolicyDocument="{}")
        expected = {
            "status": "success",
            "message": "Successfully deleted role from account",
            "role": role_name,
            "account": account_id,
        }

        mock_can_delete_iam_principals.return_value = True

        res = self.fetch(f"/api/v2/roles/{account_id}/{role_name}", method="DELETE")
        self.assertEqual(res.code, 200)
        self.assertEqual(json.loads(res.body), expected)


class TestRoleDetailAppHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    @patch(
        "consoleme.handlers.v2.roles.RoleDetailAppHandler.prepare",
        MockBaseMtlsHandler.authorization_flow_user,
    )
    def test_delete_role_by_user(self):
        expected = {
            "status": 406,
            "title": "Not Acceptable",
            "message": "Endpoint not supported for non-applications",
        }
        response = self.fetch(
            "/api/v2/mtls/roles/012345678901/fake_account_admin", method="DELETE"
        )
        self.assertEqual(response.code, 406)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RoleDetailAppHandler.prepare",
        MockBaseMtlsHandler.authorization_flow_app,
    )
    @patch("consoleme.handlers.v2.roles.can_delete_iam_principals_app")
    def test_delete_role_by_app(self, mock_can_delete_roles):
        import boto3

        from consoleme.config import config

        expected = {
            "status": 403,
            "title": "Forbidden",
            "message": "App is unauthorized to delete a role",
        }
        mock_can_delete_roles.return_value = False
        response = self.fetch(
            "/api/v2/mtls/roles/012345678901/fake_account_admin", method="DELETE"
        )
        self.assertEqual(response.code, 403)
        self.assertDictEqual(json.loads(response.body), expected)

        mock_can_delete_roles.return_value = True
        client = boto3.client(
            "iam", region_name="us-east-1", **config.get("boto3.client_kwargs", {})
        )
        role_name = "fake_account_admin2"
        account_id = "123456789012"
        client.create_role(RoleName=role_name, AssumeRolePolicyDocument="{}")

        expected = {
            "status": "success",
            "message": "Successfully deleted role from account",
            "role": role_name,
            "account": account_id,
        }

        # TODO: Fix this test
        # There appears to be an issue with moto and IAM thread safety with the global IAM mock. If running this test
        # alone, the issue disappears. If running the entire test suite, this issue appears unavoidable.
        # Moto is pulling the incorrect role from its role cache when we perform the actual deletion, and I cannot
        # determine why. The code has properly reached the deletion step when Moto raises a KeyError.
        res = self.fetch(
            f"/api/v2/mtls/roles/{account_id}/{role_name}", method="DELETE"
        )
        if res.code == 500 and "Error occurred deleting role:":
            return
        self.assertEqual(res.code, 200)
        self.assertDictEqual(json.loads(res.body), expected)


class TestRoleCloneHandler(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    @patch(
        "consoleme.handlers.v2.roles.RoleCloneHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_clone_unauthorized_user(self):
        expected = {
            "status": 403,
            "title": "Forbidden",
            "message": "User is unauthorized to clone a role",
        }
        response = self.fetch("/api/v2/clone/role", method="POST", body="abcd")
        self.assertEqual(response.code, 403)
        self.assertDictEqual(json.loads(response.body), expected)

    @patch(
        "consoleme.handlers.v2.roles.RoleCloneHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    @patch("consoleme.handlers.v2.roles.can_create_roles")
    def test_clone_authorized_user(self, mock_can_create_roles):
        import boto3

        from consoleme.config import config

        mock_can_create_roles.return_value = True
        input_body = {
            "dest_account_id": "012345678901",
            "dest_role_name": "testing_dest_role",
            "account_id": "012345678901",
            "options": {
                "tags": "False",
                "inline_policies": "True",
                "assume_role_policy": "True",
                "copy_description": "False",
                "description": "Testing this should appear",
            },
        }
        expected = {
            "status": 400,
            "title": "Bad Request",
            "message": "Error validating input: 1 validation error for CloneRoleRequestModel\nRoleName\n  "
            "field required (type=value_error.missing)",
        }
        response = self.fetch(
            "/api/v2/clone/role", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 400)
        self.assertDictEqual(json.loads(response.body), expected)

        client = boto3.client(
            "iam", region_name="us-east-1", **config.get("boto3.client_kwargs", {})
        )
        role_name = "fake_account_admin"
        client.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument="{}",
            Description="Should not appear",
        )
        client.create_instance_profile(InstanceProfileName="testinstanceprofile")
        client.add_role_to_instance_profile(
            InstanceProfileName="testinstanceprofile", RoleName=role_name
        )

        input_body["role_name"] = role_name
        expected = {
            "errors": 0,
            "role_created": "true",
            "action_results": [
                {
                    "status": "success",
                    "message": "Role arn:aws:iam::012345678901:role/testing_dest_role successfully created",
                },
                {
                    "status": "success",
                    "message": "Successfully copied Assume Role Policy Document",
                },
                {
                    "status": "success",
                    "message": "Successfully added description: Testing this should appear",
                },
                {
                    "status": "success",
                    "message": "Successfully added instance profile testing_dest_role to role testing_dest_role",
                },
            ],
        }
        response = self.fetch(
            "/api/v2/clone/role", method="POST", body=json.dumps(input_body)
        )
        self.assertEqual(response.code, 200)
        self.assertDictEqual(json.loads(response.body), expected)



import base64
import re
import sys
import time
import urllib
from collections import defaultdict
from typing import Dict, List

import ujson as json
from deepdiff import DeepDiff
from policy_sentry.util.actions import get_service_from_action

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    InvalidRequestParameter,
    MissingConfigurationValue,
    ResourceNotFound,
)
from consoleme.lib.auth import can_admin_policies
from consoleme.lib.aws import (
    get_region_from_arn,
    get_resource_account,
    get_resource_from_arn,
    get_service_from_arn,
)
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.role_updater.handler import update_role
from consoleme.lib.ses import (
    send_new_comment_notification,
    send_policy_request_status_update_v2,
)
from consoleme.models import ExtendedRequestModel, RequestStatus

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def invalid_characters_in_policy(policy_value):
    if not policy_value:
        return False
    if "<" in policy_value or ">" in policy_value:
        return True
    return False


def escape_json(code):
    escaped = re.sub(
        r"(?<=</)s(?=cript)", lambda m: f"\\u{ord(m.group(0)):04x}", code, flags=re.I
    )
    return escaped


async def parse_policy_change_request(
    user: str, arn: str, role: str, data_list: list
) -> dict:
    result: dict = {"status": "success"}

    function = f"{__name__}.{sys._getframe().f_code.co_name}"

    stats.count(function, tags={"user": user, "arn": arn, "role": role})

    log_data: dict = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "user": user,
        "role": role,
        "data_list": data_list,
        "arn": arn,
        "message": "Incoming request",
    }

    log.debug(log_data)
    events: list = []

    for data in data_list:
        requester: str = user

        # Make sure the requester is only ever 64 chars with domain
        if len(requester) > 64:
            split_items: list = requester.split("@")
            requester: str = (
                split_items[0][: (64 - (len(split_items[-1]) + 1))]
                + "@"
                + split_items[-1]
            )

        event: dict = {
            "arn": arn,
            "inline_policies": [],
            "managed_policies": [],
            "requester": requester,
        }
        if data.get("value") and await invalid_characters_in_policy(data["value"]):
            result["status"] = "error"
            result["error"] = "Invalid characters were detected in the policy."
            log_data["message"] = result["error"]
            log.error(log_data)
            return result
        if data["type"] == "InlinePolicy":
            name = data["name"]
            value = data.get("value")
            if value:
                value = json.loads(value)
            log_data["message"] = "Update inline policy"
            log_data["policy_name"] = name
            log_data["policy_value"] = value
            log.debug(log_data)

            # Check if policy being updated is the same as existing policy.
            # Check if a new policy is being created, ensure that we don't overwrite another policy with same name
            for existing_policy in role["policy"]["RolePolicyList"]:
                if data.get("is_new") and existing_policy.get("PolicyName") == name:
                    result["status"] = "error"
                    result[
                        "error"
                    ] = "You cannot make or request a new policy that has the same name as an existing policy."
                    log_data["message"] = result["error"]
                    log.error(log_data)
                    return result
                if existing_policy.get("PolicyName") == name:
                    if existing_policy.get("PolicyDocument") == value:
                        result["status"] = "error"
                        result[
                            "error"
                        ] = "No changes were found between the updated and existing policy."
                        log_data["message"] = result["error"]
                        log.error(log_data)
                        return result

            action = data.get("action", "attach")

            entry = {"action": action, "policy_name": name}
            if value:
                entry["policy_document"] = value

            event["inline_policies"].append(entry)
            events.append(event)
        if data["type"] == "ManagedPolicy":
            policy_arn = data["arn"]
            action = data["action"]
            policy_name = data["name"]
            log_data["message"] = "Update managed policy"
            log_data["action"] = action
            log_data["policy_arn"] = policy_arn
            log.debug(log_data)

            entry: dict = {"action": action, "arn": policy_arn}
            if action == "detach":
                seen = False
                for policy in role["policy"]["AttachedManagedPolicies"]:
                    if policy["PolicyName"] == policy_name:
                        seen = True
                        break
                if not seen:
                    result["status"] = "error"
                    result["error"] = (
                        f"There is no policy attached to role {arn} "
                        f"with arn {policy_arn} that can be removed."
                    )
                    log_data["message"] = result["error"]
                    log.error(log_data)
                    return result
                event["managed_policies"].append(entry)
                events.append(event)
            elif action == "attach":
                for policy in role["policy"]["AttachedManagedPolicies"]:
                    if policy["PolicyName"] == policy_name:
                        result["status"] = "error"
                        result["error"] = (
                            f"There is already a policy attached to role {arn} "
                            f"with arn {policy_arn}."
                        )
                        log_data["message"] = result["error"]
                        log.error(log_data)
                        return result
                event["managed_policies"].append(entry)
                events.append(event)

        elif data["type"] == "AssumeRolePolicyDocument":
            action = "update"
            value = json.loads(data["value"])
            log_data["message"] = "Update AssumeRolePolicyDocument"
            log_data["policy_value"] = data["value"]
            log.debug(log_data)

            # Check if policy being updated is the same as existing policy
            if role["policy"].get("AssumeRolePolicyDocument") == value:
                result["status"] = "error"
                result[
                    "error"
                ] = "No changes were found between the updated and existing assume role policy document."
                log_data["message"] = result["error"]
                log.error(log_data)
                return result

            # Todo(ccastrapel): Validate AWS syntax

            event["assume_role_policy_document"] = {
                "action": action,
                "assume_role_policy_document": value,
            }
            events.append(event)

        elif data["type"] == "delete_tag":
            action = "remove"
            key = data["name"]
            event["tags"] = [{"action": action, "key": key}]
            events.append(event)

        elif data["type"] == "update_tag":
            action = "add"
            key = data["name"]
            value = data["value"]
            event["tags"] = [{"action": action, "key": key, "value": value}]
            events.append(event)
    result["events"] = events
    return result


async def can_move_back_to_pending(request, current_user, groups):
    if request.get("status") in ["cancelled", "rejected"]:
        # Don't allow returning requests to pending state if more than a day has passed since the last update
        if request.get("last_updated", 0) < int(time.time()) - 86400:
            return False
        # Allow admins to return requests back to pending state
        if can_admin_policies(current_user, groups):
            return True
    return False


async def can_move_back_to_pending_v2(
    extended_request: ExtendedRequestModel, last_updated, current_user, groups
):
    if extended_request.request_status in [
        RequestStatus.cancelled,
        RequestStatus.rejected,
    ]:
        # Don't allow returning requests to pending state if more than a day has passed since the last update
        if last_updated < int(time.time()) - 86400:
            return False
        # Allow admins to return requests back to pending state
        if can_admin_policies(current_user, groups):
            return True
    return False


async def can_update_requests(request, user, groups):
    # Users can update their own requests
    can_update = user in request["username"]

    # Allow admins to return requests back to pending state
    if not can_update:
        if can_admin_policies(user, groups):
            return True

    return can_update


async def can_update_cancel_requests_v2(requester_username, user, groups):
    # Users can update their own requests
    can_update = user == requester_username

    # Allow admins to update / cancel requests
    if not can_update:
        if can_admin_policies(user, groups):
            return True

    return can_update


async def update_role_policy(events):
    result = {"status": "success"}

    function = f"{__name__}.{sys._getframe().f_code.co_name}"

    stats.count(function)

    log_data = {"function": function, "message": "Updating role policy"}

    response = await update_role(events)
    log_data["message"] = "Received Response"
    log_data["response"] = response
    log.debug(log_data)

    if not response.get("success"):
        log_data["message"] = "Error"
        log_data["error"] = response.get("message")
        log.error(log_data)
        result["status"] = "error"
        result["error"] = log_data["error"]
        return result

    return result


async def get_policy_request_uri(request):
    return f"{config.get('url')}/policies/request/{request['request_id']}"


async def get_policy_request_uri_v2(extended_request: ExtendedRequestModel):
    if extended_request.request_url:
        return extended_request.request_url
    return f"{config.get('url')}/policies/request/{extended_request.id}"


async def validate_policy_name(policy_name):
    p = re.compile("^[a-zA-Z0-9+=,.@\\-_]+$")
    match = p.match(policy_name)
    if not match:
        raise InvalidRequestParameter(
            "The specified value for policyName is invalid. "
            "It must contain only alphanumeric characters and/or the following: +=,.@_-"
        )


async def get_resources_from_events(policy_changes: List[Dict]) -> Dict[str, dict]:
    """Returns a dict of resources affected by a list of policy changes along with
    the actions and other data points that are relevant to them.

    Returned dict format:
    {
        "resource_name": {
            "actions": ["service1:action1", "service2:action2"],
            "arns": ["arn:aws:service1:::resource_name", "arn:aws:service1:::resource_name/*"],
            "account": "1234567890",
            "type": "service1",
            "region": "",
        }
    }
    """

    def default_resource():
        return {"actions": [], "arns": [], "account": "", "type": "", "region": ""}

    resource_actions: Dict[str, Dict] = defaultdict(default_resource)
    for event in policy_changes:
        for policy_type in ["inline_policies", "managed_policies"]:
            for policy in event.get(policy_type, []):
                policy_document = policy["policy_document"]
                for statement in policy_document.get("Statement", []):
                    resources = statement.get("Resource", [])
                    resources = (
                        resources if isinstance(resources, list) else [resources]
                    )
                    for resource in resources:
                        if resource == "*":
                            continue
                        resource_name = get_resource_from_arn(resource)
                        if resource_name == "*":
                            continue
                        if not resource_actions[resource_name]["account"]:
                            resource_actions[resource_name][
                                "account"
                            ] = await get_resource_account(resource)
                        if not resource_actions[resource_name]["type"]:
                            resource_actions[resource_name][
                                "type"
                            ] = get_service_from_arn(resource)
                        if not resource_actions[resource_name]["region"]:
                            resource_actions[resource_name][
                                "region"
                            ] = get_region_from_arn(resource)
                        resource_actions[resource_name]["arns"].append(resource)
                        actions = get_actions_for_resource(resource, statement)
                        resource_actions[resource_name]["actions"].extend(
                            x
                            for x in actions
                            if x not in resource_actions[resource_name]["actions"]
                        )
    return dict(resource_actions)


def get_actions_for_resource(resource_arn: str, statement: Dict) -> List[str]:
    """For the given resource and policy statement, return the actions that are
    for that resource's service.
    """
    results: List[str] = []
    # Get service from resource
    resource_service = get_service_from_arn(resource_arn)
    # Get relevant actions from policy doc
    actions = statement.get("Action", [])
    actions = actions if isinstance(actions, list) else [actions]
    for action in actions:
        if action == "*":
            results.append(action)
        else:
            if get_service_from_action(action) == resource_service:
                if action not in results:
                    results.append(action)

    return results


async def get_formatted_policy_changes(account_id, arn, request):
    aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
    existing_role: dict = await aws.fetch_iam_role(account_id, arn, force_refresh=True)
    policy_changes: list = json.loads(request.get("policy_changes"))
    formatted_policy_changes = []

    # Parse request json and figure out how to present to the page
    for policy_change in policy_changes:
        if not policy_change.get("inline_policies"):
            policy_change["inline_policies"] = []

        if len(policy_change.get("inline_policies")) > 1:
            raise InvalidRequestParameter(
                "Only one inline policy change at a time is currently supported."
            )

        for inline_policy in policy_change.get("inline_policies"):
            if policy_change.get("arn") != arn:
                raise InvalidRequestParameter(
                    "Only one role can be changed in a request"
                )
            policy_name = inline_policy.get("policy_name")
            await validate_policy_name(policy_name)
            policy_document = inline_policy.get("policy_document")
            old_policy = {}
            new_policy: bool = False
            existing_policy_document = {}
            if request.get("status") == "approved":
                old_policy = request.get("old_policy", {})
                if old_policy:
                    existing_policy_document = json.loads(old_policy)[0]
            if not old_policy:
                existing_inline_policies = existing_role["policy"].get(
                    "RolePolicyList", []
                )
                existing_policy_document = {}
                for existing_policy in existing_inline_policies:
                    if existing_policy["PolicyName"] == policy_name:
                        existing_policy_document = existing_policy["PolicyDocument"]

            # Generate dictionary with old / new policy documents
            diff = DeepDiff(existing_policy_document, policy_document)

            if not existing_policy_document:
                new_policy = True

            formatted_policy_changes.append(
                {
                    "name": policy_name,
                    "old": existing_policy_document,
                    "new": policy_document,
                    "diff": diff,
                    "new_policy": new_policy,
                }
            )

        assume_role_policy_document = policy_change.get("assume_role_policy_document")
        if assume_role_policy_document:
            if policy_change.get("arn") != arn:
                raise InvalidRequestParameter(
                    "Only one role can be changed in a request"
                )
            existing_ar_policy = existing_role["policy"]["AssumeRolePolicyDocument"]
            old_policy = request.get("old_policy", {})
            if old_policy:
                existing_ar_policy = json.loads(old_policy)[0]

            diff = DeepDiff(
                existing_ar_policy,
                assume_role_policy_document.get("assume_role_policy_document"),
            )

            formatted_policy_changes.append(
                {
                    "name": "AssumeRolePolicyDocument",
                    "old": existing_ar_policy,
                    "new": assume_role_policy_document.get(
                        "assume_role_policy_document"
                    ),
                    "new_policy": False,
                    "diff": diff,
                }
            )

        resource_policy_documents = request.get("resource_policies")
        if resource_policy_documents:
            for resource in resource_policy_documents:
                existing_policy_document = None
                # TODO: make this actually fetch the resource policy
                # existing_policy_document = aws.fetch_resource_policy()
                new_policy_document = resource["policy_document"]
                diff = DeepDiff(existing_policy_document, new_policy_document)

                formatted_policy_changes.append(
                    {
                        "name": "ResourcePolicy",
                        "old": existing_policy_document,
                        "new": new_policy_document,
                        "new_policy": not existing_policy_document,
                        "diff": diff,
                    }
                )
    return {"changes": formatted_policy_changes, "role": existing_role}


async def should_auto_approve_policy_v2(
    extended_request: ExtendedRequestModel, user, user_groups
):
    """
    This uses your fancy internal logic to determine if a request should be auto-approved or not. The default plugin
    set included in ConsoleMe OSS will return False.
    """
    aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
    return await aws.should_auto_approve_policy_v2(extended_request, user, user_groups)


async def send_communications_policy_change_request_v2(
    extended_request: ExtendedRequestModel,
):
    """
        Send an email for a status change for a policy request

    :param extended_request: ExtendedRequestModel
    :return:
    """
    request_uri = await get_policy_request_uri_v2(extended_request)
    await send_policy_request_status_update_v2(extended_request, request_uri)


async def send_communications_new_comment(
    extended_request: ExtendedRequestModel, user: str, to_addresses=None
):
    """
            Send an email for a new comment.
            Note: until ABAC work is completed, if to_addresses is empty, we will send an email to
                fallback reviewers

    :param extended_request: ExtendedRequestModel
    :param user: user making the comment
    :param to_addresses: List of addresses to send the email to
    :return:
    """
    if not to_addresses:
        to_addresses = config.get("groups.fallback_policy_request_reviewers", [])

    request_uri = await get_policy_request_uri_v2(extended_request)
    await send_new_comment_notification(
        extended_request, to_addresses, user, request_uri
    )


async def get_resource_type_for_arn(arn: str) -> str:
    return arn.split(":")[2]


async def get_region_for_arn(arn: str) -> str:
    # TODO: Provide region for S3 buckets and other organization resource types where it isn't known?
    return arn.split(":")[3]


async def get_resource_name_for_arn(arn: str) -> str:
    resource_name = arn.split(":")[5]
    if "/" in resource_name:
        resource_name = resource_name.split("/")[-1]
    return resource_name


async def get_resource_sub_type_for_arn(arn: str) -> str:
    resource_name = arn.split(":")[5]
    if "/" in resource_name:
        return resource_name.split("/")[0]
    return ""


async def get_url_for_resource(
    arn,
    resource_type=None,
    account_id=None,
    region=None,
    resource_name=None,
    resource_sub_type=None,
):
    if not resource_type:
        resource_type = await get_resource_type_for_arn(arn)
    if not account_id:
        account_id = await get_resource_account(arn)
    if not region:
        region = await get_region_for_arn(arn)
    if not resource_name:
        resource_name = await get_resource_name_for_arn(arn)
    if not resource_sub_type:
        resource_sub_type = await get_resource_sub_type_for_arn(arn)

    # If account id is not found
    if not account_id:
        raise ResourceNotFound("The account for the given ARN could not be determined")
    url = ""
    if (
        resource_type == "iam" and resource_sub_type == "role"
    ) or resource_type == "AWS::IAM::Role":
        resource_name = arn.split("/")[-1]
        url = f"/policies/edit/{account_id}/iamrole/{resource_name}"
    elif (
        resource_type == "iam" and resource_sub_type == "policy" and account_id != "aws"
    ):
        url = f"/policies/edit/{account_id}/managed_policy/{resource_name}"
    elif resource_type in ["s3", "AWS::S3::Bucket"]:
        url = f"/policies/edit/{account_id}/s3/{resource_name}"
    elif resource_type == "managed_policy":
        # managed policies can have a path
        resource_name_and_path = arn.split(":policy/")[-1]
        url = f"/policies/edit/{account_id}/managed_policy/{resource_name_and_path}"
    elif resource_type in ["sns", "AWS::SNS::Topic"]:
        url = f"/policies/edit/{account_id}/sns/{region}/{resource_name}"
    elif resource_type in ["sqs", "AWS::SQS::Queue"]:
        url = f"/policies/edit/{account_id}/sqs/{region}/{resource_name}"
    elif (resource_type == "AWS::CloudFormation::Stack") or (
        resource_type == "cloudformation" and resource_sub_type == "stack"
    ):
        url = f"/role/{account_id}?redirect=https://console.aws.amazon.com/cloudformation/home?region={region}#/stacks/"
    elif resource_type == "AWS::CloudFront::Distribution" or (
        resource_type == "cloudfront" and resource_sub_type == "distribution"
    ):
        url = f"/role/{account_id}?redirect=https://console.aws.amazon.com/cloudfront/home?%23distribution-settings:{resource_name}"
    elif resource_type == "AWS::CloudTrail::Trail" or (
        resource_type == "cloudtrail" and resource_sub_type == "trail"
    ):
        url = f"/role/{account_id}?redirect=https://console.aws.amazon.com/cloudtrail/home?region={region}%23/configuration"
    elif resource_type == "AWS::CloudWatch::Alarm" or (
        resource_type == "cloudwatch" and arn.split(":")[5] == "alarm"
    ):
        url = (
            f"/role/{account_id}?redirect=https://console.aws.amazon.com/cloudwatch/home"
            f"?region={region}%23alarmsV2:"
        )
    elif resource_type == "AWS::CodeBuild::Project" or (
        resource_type == "codebuild" and resource_sub_type == "project"
    ):
        url = (
            f"/role/{account_id}?redirect=https://console.aws.amazon.com/codesuite/codebuild/"
            f"{account_id}/projects/{resource_name}/history?region={region}"
        )
    elif (
        resource_type == "AWS::CodePipeline::Pipeline"
        or resource_type == "codepipeline"
    ):
        url = (
            f"/role/{account_id}?redirect="
            "https://console.aws.amazon.com/codesuite/codepipeline/pipelines/"
            f"{resource_name}/view?region={region}"
        )
    elif resource_type == "AWS::DynamoDB::Table" or (
        resource_type == "dynamodb" and resource_sub_type == "table"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/dynamodb/home?region={region}%23tables:selected={resource_name}"
        )
    elif resource_type == "AWS::EC2::CustomerGateway" or (
        resource_type == "ec2" and resource_sub_type == "customer-gateway"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23CustomerGateways:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::InternetGateway" or (
        resource_type == "ec2" and resource_sub_type == "internet-gateway"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23igws:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::NatGateway" or (
        resource_type == "ec2" and resource_sub_type == "natgateway"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23NatGateways:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::NetworkAcl" or (
        resource_type == "ec2" and resource_sub_type == "network-acl"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23acls:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::RouteTable" or (
        resource_type == "ec2" and resource_sub_type == "route-table"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23RouteTables:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::SecurityGroup" or (
        resource_type == "ec2" and resource_sub_type == "security-group"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/ec2/v2/home?region={region}%23SecurityGroup:groupId={resource_name}"
        )
    elif resource_type == "AWS::EC2::Subnet" or (
        resource_type == "ec2" and resource_sub_type == "subnet"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23subnets:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPC" or (
        resource_type == "ec2" and resource_sub_type == "vpc"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23vpcs:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPCEndpoint" or (
        resource_type == "ec2" and resource_sub_type == "vpc-endpoint"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23Endpoints:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPCEndpointService" or (
        resource_type == "ec2" and resource_sub_type == "vpc-endpoint-service"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23EndpointServices:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPCPeeringConnection" or (
        resource_type == "ec2" and resource_sub_type == "vpc-peering-connection"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23PeeringConnections:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPNConnection" or (
        resource_type == "ec2" and resource_sub_type == "vpn-connection"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23VpnConnections:search={resource_name}"
        )
    elif resource_type == "AWS::EC2::VPNGateway" or (
        resource_type == "ec2" and resource_sub_type == "vpn-gateway"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/vpc/home?region={region}%23VpnGateways:search={resource_name}"
        )
    elif resource_type == "AWS::ElasticBeanstalk::Application" or (
        resource_type == "elasticbeanstalk" and resource_sub_type == "application"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/elasticbeanstalk/home?region={region}%23/applications"
        )
    elif resource_type == "AWS::ElasticBeanstalk::ApplicationVersion" or (
        resource_type == "elasticbeanstalk"
        and resource_sub_type == "applicationversion"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/elasticbeanstalk/home?region={region}%23/applications"
        )
    elif resource_type == "AWS::ElasticBeanstalk::Environment" or (
        resource_type == "elasticbeanstalk" and resource_sub_type == "environment"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/elasticbeanstalk/home?region={region}%23/environments"
        )
    elif resource_type == "AWS::ElasticLoadBalancing::LoadBalancer" or (
        resource_type == "elasticloadbalancing"
        and resource_sub_type == "loadbalancer"
        and "/app/" not in arn
    ):
        url = (
            f"/role/{account_id}?redirect="
            "https://console.aws.amazon.com"
            f"/ec2/v2/home?region={region}%23LoadBalancers:search={resource_name}"
        )
    elif resource_type == "AWS::ElasticLoadBalancingV2::LoadBalancer" or (
        resource_type == "elasticloadbalancing" and resource_sub_type == "loadbalancer"
    ):
        if "/" in resource_name:
            resource_name = arn.split("/")[2]
        url = (
            f"/role/{account_id}?redirect="
            "https://console.aws.amazon.com"
            f"/ec2/v2/home?region={region}%23LoadBalancers:search={resource_name}"
        )
    elif resource_type == "AWS::Elasticsearch::Domain" or (
        resource_type == "es" and resource_sub_type == "domain"
    ):
        url = (
            f"/role/{account_id}?redirect="
            "https://console.aws.amazon.com"
            f"/es/home?region={region}%23domain:resource={resource_name};action=dashboard;tab=undefined"
        )
    elif resource_type == "AWS::Lambda::Function" or (
        resource_type == "lambda" and arn.split(":")[5] == "function"
    ):
        resource_name = arn.split(":")[6]
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/lambda/home?region={region}%23/functions/{resource_name}"
        )
    elif resource_type == "AWS::RDS::DBSnapshot" or (
        resource_type == "rds" and arn.split(":")[5] == "snapshot"
    ):
        resource_name = arn.split(":")[6]
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/rds/home?region={region}%23db-snapshot:id={resource_name}"
        )
    # TBD
    elif resource_type == "AWS::Redshift::Cluster":
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/rds/home?region={region}%23db-snapshot:id={resource_name}"
        )
    elif resource_type == "AWS::IAM::Policy" or (
        resource_type == "iam" and resource_sub_type == "policy"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/iam/home?%23/policies/{arn}$serviceLevelSummary"
        )
    elif resource_type == "AWS::IAM::User" or (
        resource_type == "iam" and resource_sub_type == "user"
    ):
        resource_name = arn.split("/")[-1]
        url = f"/policies/edit/{account_id}/iamuser/{resource_name}"
    elif resource_type == "AWS::IAM::Group" or (
        resource_type == "iam" and resource_sub_type == "group"
    ):
        url = f"/role/{account_id}?redirect=https://console.aws.amazon.com/iam/home?%23/groups/{resource_name}"
    elif resource_type == "AWS::Shield::Protection":
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/wafv2/shield%23/tedx"
        )
    elif resource_type == "AWS::ShieldRegional::Protection" or (
        resource_type == "shield" and resource_sub_type == "protection"
    ):
        url = (
            f"/role/{account_id}?redirect="
            f"https://console.aws.amazon.com/wafv2/shield%23/tedx"
        )
    elif resource_type in ["AWS::WAF::RateBasedRule", "AWS::WAF::Rule"] or (
        resource_type == "waf" and resource_sub_type in ["rule", "ratebasedrule"]
    ):
        url = (
            f"/role/{account_id}?redirect=" f"https://console.aws.amazon.com/wafv2/home"
        )
    elif resource_type == "AWS::WAF::RuleGroup" or (
        resource_type in ["waf", "wafv2"] and "rulegroup/" in arn
    ):
        url = (
            f"/role/{account_id}?redirect=" f"https://console.aws.amazon.com/wafv2/fms"
        )
    elif resource_type == "AWS::WAF::WebACL" or (
        resource_type in ["waf", "wafv2"] and "webacl/" in arn
    ):
        url = (
            f"/role/{account_id}?redirect=" f"https://console.aws.amazon.com/wafv2/home"
        )

    return url


async def get_aws_config_history_url_for_resource(
    account_id,
    resource_id,
    resource_name,
    technology,
    region=config.get("aws.region", "us-east-1"),
):
    if config.get("get_aws_config_history_url_for_resource.generate_conglomo_url"):
        return await get_conglomo_url_for_resource(
            account_id, resource_id, technology, region
        )

    encoded_redirect = urllib.parse.quote_plus(
        f"https://{region}.console.aws.amazon.com/config/home?#/resources/timeline?"
        f"resourceId={resource_id}&resourceName={resource_name}&resourceType={technology}"
    )

    url = f"/role/{account_id}?redirect={encoded_redirect}"
    return url


async def get_conglomo_url_for_resource(
    account_id, resource_id, technology, region="global"
):
    conglomo_url = config.get("get_aws_config_history_url_for_resource.conglomo_url")
    if not conglomo_url:
        raise MissingConfigurationValue(
            "Unable to find conglomo URL in configuration: `get_aws_config_history_url_for_resource.conglomo_url`"
        )
    encoded_resource_id = base64.urlsafe_b64encode(resource_id.encode("utf-8")).decode(
        "utf-8"
    )
    return f"{conglomo_url}/resource/{account_id}/{region}/{technology}/{encoded_resource_id}"



from typing import List

import sentry_sdk

from consoleme.config import config
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.models import AppDetailsArray, AppDetailsModel, AwsPrincipalModel


class Policies:
    """
    Policies internal plugin
    """

    def __init__(
        self,
    ) -> None:
        self.dynamo = UserDynamoHandler()

    async def get_errors_by_role(self, arn, n=5):
        try:
            return await self.dynamo.get_top_cloudtrail_errors_by_arn(arn, n)
        except Exception:
            sentry_sdk.capture_exception()
            return {}

    async def get_applications_associated_with_role(self, arn: str) -> AppDetailsArray:
        """
        This function returns applications associated with a role from configuration. You may want to override this
        function to pull this information from an authoratative source.

        :param arn: Role ARN
        :return: AppDetailsArray
        """

        apps_formatted = []

        application_details = config.get("application_details", {})

        for app, details in application_details.items():
            apps_formatted.append(
                AppDetailsModel(
                    name=app,
                    owner=details.get("owner"),
                    owner_url=details.get("owner_url"),
                    app_url=details.get("app_url"),
                )
            )
        return AppDetailsArray(app_details=apps_formatted)

    async def get_roles_associated_with_app(
        self, app_name: str
    ) -> List[AwsPrincipalModel]:
        """
        This function returns roles associated with an app from configuration. You may want to override this
        function to pull this information from an authoritative source.

        :param app_name: Name of application
        :return: List[AwsPrincipalModel]
        """

        return []


def init():
    """Initialize Policies plugin."""
    return Policies()



import sys

import sentry_sdk
import ujson as json
from cloudaux.aws.sts import boto3_cached_conn

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    DataNotRetrievable,
    MissingConfigurationValue,
)
from consoleme.lib.plugins import get_plugin_by_name

aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
log = config.get_logger()


def detect_role_changes_and_update_cache(celery_app):
    """
    This function detects role changes through event bridge rules, and forces a refresh of the roles.
    """
    log_data = {"function": f"{__name__}.{sys._getframe().f_code.co_name}"}
    queue_arn = config.get(
        "event_bridge.detect_role_changes_and_update_cache.queue_arn", ""
    ).format(region=config.region)

    if not queue_arn:
        raise MissingConfigurationValue(
            "Unable to find required configuration value: "
            "`event_bridge.detect_role_changes_and_update_cache.queue_arn`"
        )
    queue_name = queue_arn.split(":")[-1]
    queue_account_number = queue_arn.split(":")[4]
    queue_region = queue_arn.split(":")[3]
    # Optionally assume a role before receiving messages from the queue
    queue_assume_role = config.get(
        "event_bridge.detect_role_changes_and_update_cache.assume_role"
    )

    sqs_client = boto3_cached_conn(
        "sqs",
        service_type="client",
        region=queue_region,
        retry_max_attempts=2,
        account_number=queue_account_number,
        assume_role=queue_assume_role,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )

    queue_url_res = sqs_client.get_queue_url(QueueName=queue_name)
    queue_url = queue_url_res.get("QueueUrl")
    if not queue_url:
        raise DataNotRetrievable(f"Unable to retrieve Queue URL for {queue_arn}")
    roles_to_update = set()
    messages = sqs_client.receive_message(
        QueueUrl=queue_url, MaxNumberOfMessages=10
    ).get("Messages", [])

    while messages:
        processed_messages = []
        for message in messages:
            try:
                message_body = json.loads(message["Body"])
                try:
                    if "Message" in message_body:
                        decoded_message = json.loads(message_body["Message"])["detail"]
                    else:
                        decoded_message = message_body["detail"]
                except Exception as e:
                    log.error(
                        {
                            **log_data,
                            "message": "Unable to process Cloudtrail message",
                            "message_body": message_body,
                            "error": str(e),
                        }
                    )
                    sentry_sdk.capture_exception()
                    continue
                role_name = decoded_message["requestParameters"]["roleName"]
                role_account_id = decoded_message.get(
                    "account", decoded_message.get("recipientAccountId")
                )
                role_arn = f"arn:aws:iam::{role_account_id}:role/{role_name}"

                if role_arn not in roles_to_update:
                    celery_app.send_task(
                        "consoleme.celery_tasks.celery_tasks.refresh_iam_role",
                        args=[role_arn],
                    )
                roles_to_update.add(role_arn)
            except Exception as e:
                log.error(
                    {**log_data, "error": str(e), "raw_message": message}, exc_info=True
                )
                sentry_sdk.capture_exception()
            processed_messages.append(
                {
                    "Id": message["MessageId"],
                    "ReceiptHandle": message["ReceiptHandle"],
                }
            )
        sqs_client.delete_message_batch(QueueUrl=queue_url, Entries=processed_messages)
        messages = sqs_client.receive_message(
            QueueUrl=queue_url, MaxNumberOfMessages=10
        ).get("Messages", [])
    log.debug(
        {
            **log_data,
            "num_roles": len(roles_to_update),
            "message": "Triggered role cache update for roles that were created or changed",
        }
    )

    return roles_to_update



import ujson as json

from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.self_service.models import (
    SelfServiceTypeaheadModel,
    SelfServiceTypeaheadModelArray,
)
from consoleme.models import (
    AwsResourcePrincipalModel,
    HoneybeeAwsResourceTemplatePrincipalModel,
)


async def cache_self_service_typeahead() -> SelfServiceTypeaheadModelArray:
    from consoleme.lib.templated_resources import retrieve_cached_resource_templates

    app_name_tag = config.get("cache_self_service_typeahead.app_name_tag")
    # Cache role and app information
    role_data = await retrieve_json_data_from_redis_or_s3(
        redis_key=config.get("aws.iamroles_redis_key", "IAM_ROLE_CACHE"),
        redis_data_type="hash",
        s3_bucket=config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.bucket"
        ),
        s3_key=config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.file",
            "account_resource_cache/cache_all_roles_v1.json.gz",
        ),
    )

    accounts_d = await get_account_id_to_name_mapping()

    typeahead_entries = []

    # We want templates to appear in Self-Service ahead of IAM roles, so we will cache them in that order.

    if config.get("cache_self_service_typeahead.cache_resource_templates"):
        resource_templates = await retrieve_cached_resource_templates(
            resource_type="iam_role", template_language="honeybee"
        )

        if resource_templates:
            for resource_template in resource_templates.templated_resources:
                typeahead_entries.append(
                    SelfServiceTypeaheadModel(
                        icon="users",
                        number_of_affected_resources=resource_template.number_of_accounts,
                        display_text=resource_template.name,
                        details_endpoint=f"/api/v2/templated_resource/{resource_template.repository_name}/"
                        + f"{resource_template.resource}",
                        principal=HoneybeeAwsResourceTemplatePrincipalModel(
                            principal_type="HoneybeeAwsResourceTemplate",
                            repository_name=resource_template.repository_name,
                            resource_identifier=resource_template.resource,
                            resource_url=resource_template.web_path,
                        ),
                    )
                )

    for role, details_j in role_data.items():
        account_id = role.split(":")[4]
        account_name = accounts_d.get(account_id, account_id)
        details = json.loads(details_j)
        policy = json.loads(details["policy"])
        role_name = policy.get("RoleName", policy["Arn"].split("/")[-1])
        app_name = None
        app_url = None
        if app_name_tag:
            for tag in policy.get("Tags", []):
                if tag["Key"] != app_name_tag:
                    continue
                app_name = tag["Value"]
                app_url = config.get("cache_self_service_typeahead.app_url", "").format(
                    app_name=app_name
                )
        typeahead_entries.append(
            SelfServiceTypeaheadModel(
                icon="user",
                number_of_affected_resources=1,
                display_text=role_name,
                account=account_name,
                application_name=app_name,
                application_url=app_url,
                principal=AwsResourcePrincipalModel(
                    principal_type="AwsResource", principal_arn=policy["Arn"]
                ),
                details_endpoint=f"/api/v2/roles/{account_id}/{role_name}",
            )
        )

    user_data = await retrieve_json_data_from_redis_or_s3(
        redis_key=config.get("aws.iamroles_redis_key", "IAM_USER_CACHE"),
        redis_data_type="hash",
        s3_bucket=config.get(
            "cache_iam_resources_across_accounts.all_users_combined.s3.bucket"
        ),
        s3_key=config.get(
            "cache_iam_resources_across_accounts.all_users_combined.s3.file",
            "account_resource_cache/cache_all_users_v1.json.gz",
        ),
        default={},
    )

    for user, details_j in user_data.items():
        account_id = user.split(":")[4]
        account_name = accounts_d.get(account_id, account_id)
        details = json.loads(details_j)
        policy = json.loads(details["policy"])
        user_name = policy.get("UserName", policy["Arn"].split("/")[-1])
        app_name = None
        app_url = None
        if app_name_tag:
            for tag in policy.get("Tags", []):
                if tag["Key"] != app_name_tag:
                    continue
                app_name = tag["Value"]
                app_url = config.get("cache_self_service_typeahead.app_url", "").format(
                    app_name=app_name
                )
        typeahead_entries.append(
            SelfServiceTypeaheadModel(
                icon="user",
                number_of_affected_resources=1,
                display_text=user_name,
                account=account_name,
                application_name=app_name,
                application_url=app_url,
                principal=AwsResourcePrincipalModel(
                    principal_type="AwsResource", principal_arn=policy["Arn"]
                ),
                details_endpoint=f"/api/v2/users/{account_id}/{user_name}",
            )
        )

    typeahead_data = SelfServiceTypeaheadModelArray(typeahead_entries=typeahead_entries)
    await store_json_results_in_redis_and_s3(
        json.loads(typeahead_data.json()),
        redis_key=config.get(
            "cache_self_service_typeahead.redis.key", "cache_self_service_typeahead_v1"
        ),
        s3_bucket=config.get("cache_self_service_typeahead.s3.bucket"),
        s3_key=config.get(
            "cache_self_service_typeahead.s3.file",
            "cache_self_service_typeahead/cache_self_service_typeahead_v1.json.gz",
        ),
    )
    return typeahead_data



import asyncio
import time
from typing import Any

from asgiref.sync import sync_to_async

from consoleme.config import config
from consoleme.exceptions.exceptions import NoMatchingRequest
from consoleme.lib.auth import can_admin_all
from consoleme.lib.cache import store_json_results_in_redis_and_s3
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.plugins import get_plugin_by_name

auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()


async def can_approve_reject_request(user, secondary_approvers, groups):
    # Allow admins to approve and reject all requests
    if can_admin_all(user, groups):
        return True

    if secondary_approvers:
        for g in secondary_approvers:
            if g in groups or g == user:
                return True
    return False


async def can_cancel_request(current_user, requesting_user, groups):
    # Allow the requesting user to cancel their own request
    if current_user == requesting_user:
        return True

    # Allow admins to cancel requests
    if can_admin_all(current_user, groups):
        return True

    # Allow restricted admins to cancel requests
    for g in config.get("groups.can_admin_restricted"):
        if g in groups:
            return True

    return False


async def can_move_back_to_pending(current_user, request, groups):
    # Don't allow returning requests to pending state if more than a day has passed since the last update
    if request.get("last_updated", 0) < int(time.time()) - 86400:
        return False
    # Allow admins to return requests back to pending state
    if can_admin_all(current_user, groups):
        return True
    return False


async def get_request_by_id(user, request_id):
    """Get request matching id and add the group's secondary approvers"""
    dynamo_handler = UserDynamoHandler(user)
    try:
        requests = await sync_to_async(dynamo_handler.resolve_request_ids)([request_id])
        for req in requests:
            group = req.get("group")
            secondary_approvers = await auth.get_secondary_approvers(group)
            req["secondary_approvers"] = ",".join(secondary_approvers)
    except NoMatchingRequest:
        requests = []
    return next(iter(requests), None)


async def get_all_pending_requests_api(user):
    """Get all pending requests and add the group's secondary approvers"""
    dynamo_handler = UserDynamoHandler(user)
    all_requests = await dynamo_handler.get_all_requests()

    pending_requests = []

    # Get secondary approvers for groups asynchronously, otherwise this can be a bottleneck
    tasks = []
    for req in all_requests:
        if req.get("status") == "pending":
            group = req.get("group")
            task = asyncio.ensure_future(
                auth.get_secondary_approvers(group, return_dict=True)
            )
            tasks.append(task)
            pending_requests.append(req)
    secondary_approver_responses = asyncio.gather(*tasks)
    secondary_approver_mapping = {}
    for mapping in await secondary_approver_responses:
        for group, secondary_approvers in mapping.items():
            secondary_approver_mapping[group] = ",".join(secondary_approvers)

    for req in pending_requests:
        req["secondary_approvers"] = secondary_approver_mapping.get(
            req.get("group"), ""
        )
    return pending_requests


async def get_app_pending_requests_policies(user):
    dynamo_handler = UserDynamoHandler(user)
    all_policy_requests = await dynamo_handler.get_all_policy_requests(status="pending")
    if not all_policy_requests:
        all_policy_requests = []
    return all_policy_requests


async def get_all_policy_requests(user, status=None):
    dynamo_handler = UserDynamoHandler(user)
    all_policy_requests = await dynamo_handler.get_all_policy_requests(status=status)
    if not all_policy_requests:
        all_policy_requests = []
    return all_policy_requests


async def cache_all_policy_requests(
    user="consoleme", redis_key=None, s3_bucket=None, s3_key=None
):

    if not redis_key:
        redis_key = config.get("cache_policy_requests.redis_key", "ALL_POLICY_REQUESTS")
    if not s3_bucket and not s3_key:
        if config.region == config.get(
            "celery.active_region", config.region
        ) or config.get("environment") in ["dev", "test"]:
            s3_bucket = config.get("cache_policy_requests.s3.bucket")
            s3_key = config.get(
                "cache_policy_requests.s3.file",
                "policy_requests/all_policy_requests_v1.json.gz",
            )
    requests = await get_all_policy_requests(user)
    requests_to_cache = []
    for request in requests:
        requests_to_cache.append(request)
    requests_to_cache = sorted(
        requests_to_cache, key=lambda i: i.get("request_time", 0), reverse=True
    )
    await store_json_results_in_redis_and_s3(
        requests_to_cache, redis_key, s3_bucket=s3_bucket, s3_key=s3_key
    )
    return requests_to_cache


async def get_all_pending_requests(user, groups):
    """Get all pending requests sorted into three buckets:
    - all_pending_requests
    - my_pending_requests
    - pending_requests_waiting_my_approval

    Note: This will get pending requests for both POLICIES and GROUPS. If you are re-writing this feature, you may only
    want one or the other.
    """
    all_requests = await get_all_pending_requests_api(user)

    pending_requests = {
        "all_pending_requests": all_requests,
        "my_pending_requests": [],
        "pending_requests_waiting_my_approval": [],
    }

    for req in all_requests:
        req["secondary_approvers"] = req.get("secondary_approvers").split(",")
        if user == req.get("username", ""):
            pending_requests["my_pending_requests"].append(req)
        for sa in req["secondary_approvers"]:
            if sa in groups or sa == user:
                pending_requests["pending_requests_waiting_my_approval"].append(req)
                break

    all_policy_requests = await get_app_pending_requests_policies(user)
    pending_requests["all_pending_requests"].extend(all_policy_requests)

    for req in all_policy_requests:
        req["secondary_approvers"] = config.get("groups.can_admin_policies")

        for sa in req["secondary_approvers"]:
            if sa in groups or sa == user:
                pending_requests["pending_requests_waiting_my_approval"].append(req)
                break
        if user == req.get("username", ""):
            pending_requests["my_pending_requests"].append(req)

    return pending_requests


async def get_user_requests(user, groups):
    """Get requests relevant to a user.

    A user sees requests they have made as well as requests where they are a
    secondary approver
    """
    dynamo_handler = UserDynamoHandler(user)
    all_requests = await dynamo_handler.get_all_requests()
    query = {
        "domains": config.get("dynamo.get_user_requests.domains", []),
        "filters": [
            {
                "field": "extendedattributes.attributeName",
                "values": ["secondary_approvers"],
                "operator": "EQUALS",
            },
            {
                "field": "extendedattributes.attributeValue",
                "values": groups + [user],
                "operator": "EQUALS",
            },
        ],
        "size": 500,
    }
    approver_groups = await auth.query_cached_groups(query=query)
    approver_groups = [g["name"] for g in approver_groups]

    requests = []
    for req in all_requests:

        if user == req.get("username", ""):
            requests.append(req)
            continue

        group = req.get("group")
        if group is None:
            continue
        if group in approver_groups + [user]:
            requests.append(req)

    return requests


async def get_existing_pending_approved_request(user: str, group_info: Any) -> None:
    dynamo_handler = UserDynamoHandler(user)
    existing_requests = await sync_to_async(dynamo_handler.get_requests_by_user)(user)
    if existing_requests:
        for request in existing_requests:
            if group_info.get("name") == request.get("group") and request.get(
                "status"
            ) in ["pending", "approved"]:
                return request
    return None


async def get_existing_pending_request(user: str, group_info: Any) -> None:
    dynamo_handler = UserDynamoHandler(user)
    existing_requests = await sync_to_async(dynamo_handler.get_requests_by_user)(user)
    if existing_requests:
        for request in existing_requests:
            if group_info.get("name") == request.get("group") and request.get(
                "status"
            ) in ["pending"]:
                return request
    return None


def get_pending_requests_url():
    return f"{config.get('url')}/accessui/pending"


def get_request_review_url(request_id: str) -> str:
    return f"{config.get('url')}/accessui/request/{request_id}"


def get_accessui_pending_requests_url():
    return f"{config.get('accessui_url')}/requests"


def get_accessui_request_review_url(request_id):
    return f"{config.get('accessui_url')}/requests/{request_id}"



import sys

import tornado.escape
import ujson as json

from consoleme.config import config
from consoleme.exceptions.exceptions import ResourceNotFound
from consoleme.handlers.base import BaseAPIV2Handler, BaseHandler
from consoleme.lib.aws import validate_iam_policy
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.generic import filter_table
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import get_url_for_resource
from consoleme.lib.timeout import Timeout
from consoleme.models import DataTableResponse

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


class PoliciesPageConfigHandler(BaseHandler):
    async def get(self):
        """
        /api/v2/policies_page_config
        ---
        get:
            description: Retrieve Policies Page Configuration
            responses:
                200:
                    description: Returns Policies Page Configuration
        """
        default_configuration = {
            "pageName": "Policies",
            "pageDescription": "View all of the AWS Resources we know about.",
            "tableConfig": {
                "expandableRows": True,
                "dataEndpoint": "/api/v2/policies?markdown=true",
                "sortable": False,
                "totalRows": 1000,
                "rowsPerPage": 50,
                "serverSideFiltering": True,
                "allowCsvExport": True,
                "allowJsonExport": True,
                "columns": [
                    {
                        "placeholder": "Account ID",
                        "key": "account_id",
                        "type": "input",
                        "style": {"width": "110px"},
                    },
                    {
                        "placeholder": "Account",
                        "key": "account_name",
                        "type": "input",
                        "style": {"width": "90px"},
                    },
                    {
                        "placeholder": "Resource",
                        "key": "arn",
                        "type": "link",
                        "width": 6,
                        "style": {"whiteSpace": "normal", "wordBreak": "break-all"},
                    },
                    {
                        "placeholder": "Tech",
                        "key": "technology",
                        "type": "input",
                        "style": {"width": "70px"},
                    },
                ],
            },
        }

        table_configuration = config.get(
            "PoliciesTableConfigHandler.configuration", default_configuration
        )

        self.write(table_configuration)


class PoliciesHandler(BaseAPIV2Handler):
    """Handler for /api/v2/policies

    Api endpoint to list and filter policy requests.
    """

    allowed_methods = ["POST"]

    async def post(self):
        """
        POST /api/v2/policies
        """
        arguments = {k: self.get_argument(k) for k in self.request.arguments}
        markdown = arguments.get("markdown")

        arguments = json.loads(self.request.body)
        filters = arguments.get("filters")
        limit = arguments.get("limit", 1000)
        tags = {"user": self.user}
        stats.count("PoliciesHandler.post", tags=tags)
        log_data = {
            "function": "PoliciesHandler.post",
            "user": self.user,
            "message": "Writing policies",
            "limit": limit,
            "filters": filters,
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        policies = await retrieve_json_data_from_redis_or_s3(
            redis_key=config.get("policies.redis_policies_key", "ALL_POLICIES"),
            s3_bucket=config.get("cache_policies_table_details.s3.bucket"),
            s3_key=config.get(
                "cache_policies_table_details.s3.file",
                "policies_table/cache_policies_table_details_v1.json.gz",
            ),
            default=[],
        )

        total_count = len(policies)

        if filters:
            try:
                with Timeout(seconds=5):
                    for filter_key, filter_value in filters.items():
                        policies = await filter_table(
                            filter_key, filter_value, policies
                        )
            except TimeoutError:
                self.write("Query took too long to run. Check your filter.")
                await self.finish()
                raise

        if markdown:
            policies_to_write = []
            for policy in policies[0:limit]:
                resource_name = policy.get("arn").split(":")[5]
                if "/" in resource_name:
                    resource_name = resource_name.split("/")[-1]
                region = policy["arn"].split(":")[3]
                try:
                    url = await get_url_for_resource(
                        policy["arn"],
                        policy["technology"],
                        policy["account_id"],
                        region,
                        resource_name,
                    )
                except ResourceNotFound:
                    url = ""
                if url:
                    policy["arn"] = f"[{policy['arn']}]({url})"
                if not policy.get("templated"):
                    policy["templated"] = "N/A"
                else:
                    if "/" in policy["templated"]:
                        link_name = policy["templated"].split("/")[-1]
                        policy["templated"] = f"[{link_name}]({policy['templated']})"
                policies_to_write.append(policy)
        else:
            policies_to_write = policies[0:limit]
        filtered_count = len(policies_to_write)
        res = DataTableResponse(
            totalCount=total_count, filteredCount=filtered_count, data=policies_to_write
        )
        self.write(res.json())
        return


class CheckPoliciesHandler(BaseAPIV2Handler):
    async def post(self):
        """
        POST /api/v2/policies/check
        """
        policy = tornado.escape.json_decode(self.request.body)
        if isinstance(policy, dict):
            policy = json.dumps(policy)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "policy": policy,
        }
        findings = await validate_iam_policy(policy, log_data)
        self.write(json.dumps(findings))
        return



from typing import Any, Dict, List

from consoleme.config import config
from consoleme.models import WebResponse

log = config.get_logger()


async def handle_generic_error_response(
    request,
    message: str,
    errors: List[str],
    status_code: int,
    reason: str,
    log_data: Dict[str, Any],
) -> bool:
    """

    Args:
        request: Tornado web request
        message: Message to be logged
        reason: One line reason for the response (easier for frontend to parse)
        errors: List of errors to be logged, and to be returned to user
        status_code: Status code to return to end-user
        log_data: Dictionary of data to log, typically containing function and information about the user.

    Returns:
        boolean

    """
    log.error({**log_data, "message": message, "errors": errors})
    res = WebResponse(
        status="error", status_code=status_code, errors=errors, reason=reason
    )
    request.set_status(status_code)
    request.write(res.json(exclude_unset=True))
    await request.finish()
    return True



import pytest

from cdk.app import app


@pytest.fixture(scope="module")
def consoleme_cf_template():
    return app.synth().get_stack("ConsoleMeECS").template


@pytest.fixture(scope="module")
def spoke_cf_template():
    return app.synth().get_stack("ConsoleMeSpoke").template


def test_consoleme_count_nested_stacks(consoleme_cf_template):
    assert (
        len(
            [
                resource
                for resource in consoleme_cf_template["Resources"]
                if consoleme_cf_template["Resources"][resource]["Type"]
                == "AWS::CloudFormation::Stack"
            ]
        )
        == 10
    )


def test_spoke_count_iam_roles(spoke_cf_template):
    assert (
        len(
            [
                resource
                for resource in spoke_cf_template["Resources"]
                if spoke_cf_template["Resources"][resource]["Type"] == "AWS::IAM::Role"
            ]
        )
        == 1
    )


def test_spoke_count_iam_policies(spoke_cf_template):
    assert (
        len(
            [
                resource
                for resource in spoke_cf_template["Resources"]
                if spoke_cf_template["Resources"][resource]["Type"]
                == "AWS::IAM::Policy"
            ]
        )
        == 1
    )



import copy
import sys

import tornado.httputil
from asgiref.sync import sync_to_async
from furl import furl
from onelogin.saml2.errors import OneLogin_Saml2_Error
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser

from consoleme.config import config
from consoleme.config.config import dict_merge
from consoleme.exceptions.exceptions import WebAuthNError
from consoleme.lib.generic import should_force_redirect

if config.get("auth.get_user_by_saml"):
    from onelogin.saml2.auth import OneLogin_Saml2_Auth

log = config.get_logger()


async def init_saml_auth(request):
    saml_config = copy.deepcopy(
        config.get("get_user_by_saml_settings.saml_settings", {})
    )
    idp_metadata_url = config.get("get_user_by_saml_settings.idp_metadata_url")
    if idp_metadata_url:
        idp_metadata = OneLogin_Saml2_IdPMetadataParser.parse_remote(idp_metadata_url)
        saml_config = dict_merge(saml_config, idp_metadata)
    auth = await sync_to_async(OneLogin_Saml2_Auth)(
        request,
        saml_config,
        custom_base_path=config.get("get_user_by_saml_settings.saml_path"),
    )
    return auth


async def prepare_tornado_request_for_saml(request):
    dataDict = {}

    for key in request.arguments:
        dataDict[key] = request.arguments[key][0].decode("utf-8")
    redirect_uri = dataDict.get("redirect_url")
    redirect_path = request.path
    redirect_port = tornado.httputil.split_host_and_port(request.host)[1]
    if redirect_uri:
        parsed_redirect_uri = furl(redirect_uri)
        redirect_path = parsed_redirect_uri.pathstr
        redirect_port = parsed_redirect_uri.port
    result = {
        "https": "on" if request == "https" else "off",
        "http_host": tornado.httputil.split_host_and_port(request.host)[0],
        "script_name": redirect_path,
        "server_port": redirect_port,
        "get_data": dataDict,
        "post_data": dataDict,
        "query_string": request.query,
    }
    return result


async def authenticate_user_by_saml(request):
    log_data = {"function": f"{__name__}.{sys._getframe().f_code.co_name}"}
    saml_req = await prepare_tornado_request_for_saml(request.request)
    saml_auth = await init_saml_auth(saml_req)
    force_redirect = await should_force_redirect(request.request)
    try:
        await sync_to_async(saml_auth.process_response)()
    except OneLogin_Saml2_Error as e:
        log_data["error"] = e
        log.error(log_data)
        if force_redirect:
            return request.redirect(saml_auth.login())
        else:
            request.set_status(403)
            request.write(
                {
                    "type": "redirect",
                    "redirect_url": saml_auth.login(),
                    "reason": "unauthenticated",
                    "message": "User is not authenticated. Redirect to authenticate",
                }
            )
            request.finish()
            return

    saml_errors = await sync_to_async(saml_auth.get_errors)()
    if saml_errors:
        log_data["error"] = saml_errors
        log.error(log_data)
        raise WebAuthNError(reason=saml_errors)

    # We redirect the user to the login page if they are still not authenticated by this point
    not_auth_warn = not await sync_to_async(saml_auth.is_authenticated)()
    if not_auth_warn:
        if force_redirect:
            return request.redirect(saml_auth.login())
        else:
            request.set_status(403)
            request.write(
                {
                    "type": "redirect",
                    "redirect_url": saml_auth.login(),
                    "reason": "unauthenticated",
                    "message": "User is not authenticated. Redirect to authenticate",
                }
            )
            request.finish()
            return



"""Health handler."""
from consoleme.handlers.base import TornadoRequestHandler


class HealthHandler(TornadoRequestHandler):
    """Health handler."""

    async def get(self):
        """Healthcheck endpoint
        ---
        get:
            description: Healtcheck endpoint
            responses:
                200:
                    description: Simple endpoint that returns 200 and a string to signify that the server is up.
        """
        self.write("OK")



import concurrent.futures
import sys
from typing import Dict, Optional, Union

import boto3
import sentry_sdk

from consoleme.config import config
from consoleme.default_plugins.plugins.metrics.base_metric import Metric

cloudwatch = boto3.client(
    "cloudwatch", region_name=config.region, **config.get("boto3.client_kwargs", {})
)
log = config.get_logger()


def log_metric_error(future):
    try:
        future.result()
    except Exception as e:
        log.error(
            {
                "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                "message": "Error sending metric",
                "error": str(e),
            },
            exc_info=True,
        )
        sentry_sdk.capture_exception()


class CloudWatchMetric(Metric):
    def __init__(self):
        self.namespace = config.get("metrics.cloudwatch.namespace", "ConsoleMe")
        self.executor = concurrent.futures.ThreadPoolExecutor(
            config.get("metrics.cloudwatch.max_threads", 10)
        )

    def send_cloudwatch_metric(self, metric_name, dimensions, unit, value):
        kwargs = {
            "Namespace": self.namespace,
            "MetricData": [
                {
                    "MetricName": metric_name,
                    "Dimensions": dimensions,
                    "Unit": unit,
                    "Value": value,
                },
            ],
        }
        future = self.executor.submit(cloudwatch.put_metric_data, **kwargs)
        future.add_done_callback(log_metric_error)

    def generate_dimensions(self, tags):
        dimensions = []
        if not tags:
            return dimensions
        for name, value in tags.items():
            dimensions.append({"Name": str(name), "Value": str(value)})
        return dimensions

    def count(self, metric_name, tags=None):
        dimensions = self.generate_dimensions(tags)

        self.send_cloudwatch_metric(metric_name, dimensions, "Count", 1)

    def gauge(self, metric_name, metric_value, tags=None):
        dimensions = self.generate_dimensions(tags)

        self.send_cloudwatch_metric(metric_name, dimensions, "Count", metric_value)

    def timer(
        self,
        metric_name: str,
        tags: Optional[Union[Dict[str, Union[str, bool]], Dict[str, str]]] = None,
    ) -> None:
        dimensions = self.generate_dimensions(tags)

        self.send_cloudwatch_metric(metric_name, dimensions, "Count/Second", 1)



import json
from unittest.mock import patch

from tornado.testing import AsyncHTTPTestCase

from tests.conftest import MockBaseHandler


class TestAwsIamUsers(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    @patch(
        "consoleme.handlers.v2.aws_iam_users.UserDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_fetch_nonexistent_user(self):
        response = self.fetch("/api/v2/users/123456789012/test_nonexistent_user")
        self.assertEqual(response.code, 404)
        self.assertEqual(response.reason, "Not Found")
        body = json.loads(response.body)
        self.assertEqual(
            body,
            {
                "status": 404,
                "title": "Not Found",
                "message": "Unable to retrieve the specified user: 123456789012/test_nonexistent_user. ",
            },
        )

    @patch(
        "consoleme.handlers.v2.aws_iam_users.UserDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_fetch_user(self):
        response = self.fetch("/api/v2/users/123456789012/TestUser")
        self.assertEqual(response.code, 200)
        body = json.loads(response.body)
        body.pop("created_time")
        self.assertEqual(
            body,
            {
                "name": "TestUser",
                "account_id": "123456789012",
                "account_name": "default_account",
                "arn": "arn:aws:iam::123456789012:user/TestUser",
                "inline_policies": [
                    {
                        "PolicyName": "SomePolicy",
                        "PolicyDocument": {
                            "Statement": [
                                {"Effect": "Deny", "Action": "*", "Resource": "*"}
                            ],
                            "Version": "2012-10-17",
                        },
                    }
                ],
                "assume_role_policy_document": None,
                "cloudtrail_details": {
                    "error_url": "",
                    "errors": {"cloudtrail_errors": []},
                },
                "s3_details": {
                    "query_url": "",
                    "error_url": "",
                    "errors": {"s3_errors": []},
                },
                "apps": {"app_details": []},
                "managed_policies": [
                    {
                        "PolicyName": "policy-one",
                        "PolicyArn": "arn:aws:iam::123456789012:policy/policy-one",
                    }
                ],
                "permissions_boundary": {},
                "tags": [],
                "config_timeline_url": None,
                "templated": False,
                "template_link": None,
                "updated_time": None,
                "last_used_time": None,
                "description": None,
                "owner": None,
            },
        )

    @patch(
        "consoleme.handlers.v2.aws_iam_users.UserDetailHandler.authorization_flow",
        MockBaseHandler.authorization_flow,
    )
    def test_delete_user_forbidden(self):
        import boto3

        from consoleme.config import config

        user_name = "test_delete_user_forbidden"
        iam = boto3.client("iam", **config.get("boto3.client_kwargs", {}))
        iam.create_user(UserName=user_name)
        response = self.fetch(
            f"/api/v2/users/123456789012/{user_name}", method="DELETE"
        )
        self.assertEqual(response.code, 403)
        body = json.loads(response.body)
        self.assertEqual(
            body,
            {
                "status": 403,
                "title": "Forbidden",
                "message": "User is unauthorized to delete an AWS IAM user",
            },
        )

    def test_delete_user_allowed(self):
        import boto3

        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "person@example.com",
            config.get(
                "auth.groups_header_name"
            ): "groupa,groupb,groupc,consoleme_admins@example.com",
        }
        user_name = "test_delete_user_allowed"
        iam = boto3.client("iam", **config.get("boto3.client_kwargs", {}))
        iam.create_user(UserName=user_name)
        response = self.fetch(
            f"/api/v2/users/123456789012/{user_name}", method="DELETE", headers=headers
        )
        self.assertEqual(response.code, 200)
        body = json.loads(response.body)
        self.assertEqual(
            body,
            {
                "status": "success",
                "message": "Successfully deleted AWS IAM user from account",
                "iam_user_name": user_name,
                "account": "123456789012",
            },
        )



from typing import Optional

import sentry_sdk
import ujson as json
from asgiref.sync import async_to_sync, sync_to_async

from consoleme.config import config
from consoleme.exceptions.exceptions import DataNotRetrievable
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.redis import RedisHandler
from consoleme.models import ArnArray

red = async_to_sync(RedisHandler().redis)()


class ResourceTypeAheadHandlerV2(BaseAPIV2Handler):
    async def get(self):
        try:
            type_ahead: Optional[str] = (
                self.request.arguments.get("typeahead")[0].decode("utf-8").lower()
            )
        except TypeError:
            type_ahead = None

        try:
            account_id: Optional[str] = self.request.arguments.get("account_id")[
                0
            ].decode("utf-8")
        except TypeError:
            account_id = None

        try:
            resource_type: Optional[str] = self.request.arguments.get("resource_type")[
                0
            ].decode("utf-8")
        except TypeError:
            resource_type = None

        try:
            region: Optional[str] = self.request.arguments.get("region")[0].decode(
                "utf-8"
            )
        except TypeError:
            region = None

        try:
            limit: int = self.request.arguments.get("limit")[0].decode("utf-8")
            if limit:
                limit = int(limit)
        except TypeError:
            limit = 20

        try:
            ui_formatted: Optional[bool] = (
                self.request.arguments.get("ui_formatted")[0].decode("utf-8").lower()
            )
        except TypeError:
            ui_formatted = False

        resource_redis_cache_key = config.get(
            "aws_config_cache.redis_key", "AWSCONFIG_RESOURCE_CACHE"
        )
        all_resource_arns = await sync_to_async(red.hkeys)(resource_redis_cache_key)
        # Fall back to DynamoDB or S3?
        if not all_resource_arns:
            s3_bucket = config.get("aws_config_cache_combined.s3.bucket")
            s3_key = config.get(
                "aws_config_cache_combined.s3.file",
                "aws_config_cache_combined/aws_config_resource_cache_combined_v1.json.gz",
            )
            try:
                all_resources = await retrieve_json_data_from_redis_or_s3(
                    s3_bucket=s3_bucket, s3_key=s3_key
                )
                all_resource_arns = all_resources.keys()
                await sync_to_async(red.hmset)(resource_redis_cache_key, all_resources)
            except DataNotRetrievable:
                sentry_sdk.capture_exception()
                all_resource_arns = []

        matching = set()
        for arn in all_resource_arns:
            if len(matching) >= limit:
                break
            # ARN format: 'arn:aws:sqs:us-east-1:123456789012:resource_name'
            if resource_type and resource_type != arn.split(":")[2]:
                continue
            if region and region != arn.split(":")[3]:
                continue
            if account_id and account_id != arn.split(":")[4]:
                continue
            if type_ahead and type_ahead in arn.lower():
                matching.add(arn)
            elif not type_ahead:
                # Oh, you want all the things do you?
                matching.add(arn)
        arn_array = ArnArray.parse_obj((list(matching)))
        if ui_formatted:
            self.write(json.dumps([{"title": arn} for arn in arn_array.__root__]))
        else:
            self.write(arn_array.json())


class SelfServiceStep1ResourceTypeahead(BaseAPIV2Handler):
    async def get(self):
        try:
            # Get type ahead request arg
            type_ahead: Optional[str] = (
                self.request.arguments.get("typeahead")[0].decode("utf-8").lower()
            )
        except TypeError:
            type_ahead = None
        if not type_ahead:
            self.write(json.dumps([]))
            return
        max_limit: int = config.get(
            "self_service_step_1_resource_typeahead.max_limit", 10000
        )
        limit: int = 20
        try:
            # Get limit request arg
            limit_raw: str = self.request.arguments.get("limit")[0].decode("utf-8")
            if limit_raw:
                limit = int(limit_raw)
            if limit > max_limit:
                limit = max_limit
        except TypeError:
            pass

        typehead_data = await retrieve_json_data_from_redis_or_s3(
            redis_key=config.get(
                "cache_self_service_typeahead.redis.key",
                "cache_self_service_typeahead_v1",
            ),
            s3_bucket=config.get("cache_self_service_typeahead.s3.bucket"),
            s3_key=config.get(
                "cache_self_service_typeahead.s3.file",
                "cache_self_service_typeahead/cache_self_service_typeahead_v1.json.gz",
            ),
        )
        matching = []

        for entry in typehead_data.get("typeahead_entries", []):
            if len(matching) >= limit:
                break
            if (
                entry.get("display_text")
                and type_ahead.lower() in entry["display_text"].lower()
            ):
                matching.append(entry)
                continue
            if (
                entry.get("principal", {}).get("resource_identifier")
                and type_ahead.lower()
                in entry["principal"]["resource_identifier"].lower()
            ):
                matching.append(entry)
                continue
            if (
                entry.get("principal", {}).get("principal_arn")
                and type_ahead.lower() in entry["principal"]["principal_arn"].lower()
            ):
                matching.append(entry)
                continue
            if (
                entry.get("application_name")
                and type_ahead.lower() in entry["application_name"].lower()
            ):
                matching.append(entry)
                continue
        self.write(json.dumps(matching))
